/*
 * Moab.c
 *
 * Copyright (c) 2007 Landon Fuller <landonf@opendarwin.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Landon Fuller nor the names of any contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */


#include <CoreFoundation/CoreFoundation.h>
#include <ApplicationEnhancer/ApplicationEnhancer.h>

#include <unistd.h>
#include <regex.h>

#include "Moab.h"
#include "MoabShared.h"

#include "javaicc_handler.h"
#include "javaicc_handler_14.h"

/*
 * Locate a mach_header using a regular expression.
 * Provided by Rosyna of Unsanity.
 */
struct mach_header *MOABGetImageHeader(const char *regex, const char **imageName) {
    unsigned long                count,index;
    struct mach_header        *header = NULL;
    const char *path;
    regex_t                                rx;

    regcomp(&rx,regex, REG_NOSUB | REG_EXTENDED);
    count = _dyld_image_count();

    for (index = 0;(header == NULL) && (index < count);index += 1) {
        path = _dyld_get_image_name(index);
        if (!regexec(&rx, path, 0, NULL, 0)) {
            header = (struct mach_header*)_dyld_get_image_header(index);
            if (imageName)
                *imageName = path;
        }
    }

    regfree(&rx);
    return header;
}

/*
 * Given an image path and a parent depth, find the enclosing CFBundle. It is the caller's
 * responsibility to release the returned bundle.
 */
CFBundleRef MOABCreateBundleForImagePathDepth(const char *imagePath, int depth) {
    CFURLRef parent, url;
    CFBundleRef bundle;
    int i;

    url = CFURLCreateFromFileSystemRepresentation(NULL, (UInt8*) imagePath, strlen(imagePath), false);

    for (i = 0; i < depth; i++) {
        if (!url) {
            return NULL;
        }

        parent = CFURLCreateCopyDeletingLastPathComponent(NULL, url);
        CFRelease(url);
        if (!parent) {
            return NULL;
        }
        url = parent;
    }

    if (!url) {
        return NULL;
    }

    bundle = CFBundleCreate(NULL, url);
    CFRelease(url);
    return bundle;
}

/*
 * Given an image path and depth, find the enclosing CFBundle and return true if the CFBundleVersion
 * matches the supplied requiredVersion argument.
 */
bool MOABMatchVersionForImageBundle(const char *imagePath, int bundleDepth, CFStringRef requiredVersion) {
    CFBundleRef bundle;
    CFDictionaryRef bundleInfoDict;
    CFStringRef versionString;

    bundle = MOABCreateBundleForImagePathDepth(imagePath, bundleDepth);
    if (!bundle) {
        apeprintf("[MOAB] Could not locate bundle for image: %s\n", imagePath);
        return false;
    }

    bundleInfoDict = CFBundleGetInfoDictionary(bundle);
    if (bundleInfoDict == NULL) {
        apeprintf("[MOAB] Could not load bundle info dictionary for image: %s\n");
        CFRelease(bundle);
        return false;
    }

    versionString = CFDictionaryGetValue(bundleInfoDict, CFSTR("CFBundleVersion"));
    if (!versionString) {
        /* No CFBundleVersion, try also CFBundleShortVersionString */
        versionString = CFDictionaryGetValue(bundleInfoDict, CFSTR("CFBundleShortVersionString"));
    }

    if (versionString && CFEqual(versionString, requiredVersion)) {
        CFRelease(bundle);
        return true;
    }

    CFRelease(bundle);
    return false;
}

bool MOABMatchVersionForApplicationBundle(CFStringRef requiredAppId, CFStringRef requiredVersion) {
    CFBundleRef appBundle;
    CFStringRef appBundleId, appVersion;

    appBundle = CFBundleGetMainBundle();
    if (!appBundle) {
        apeprintf("[MOAB] Could not locate main application bundle.\n");
        return false;
    }

    appBundleId = CFBundleGetIdentifier(appBundle);
    if (!appBundleId || !CFEqual(appBundleId, requiredAppId)) {
        return false;
    }

    appVersion = CFBundleGetValueForInfoDictionaryKey(appBundle, CFSTR("CFBundleVersion"));
    if (!appVersion || !CFEqual(appVersion, requiredVersion)) {
        return false;
    }

    return true;
}

/* APE Entry Point */
void APEBundleMainEarlyLoad(CFBundleRef bundle, CFStringRef inAPEToolsApplicationID) {
    /* 
     * CVE-2007-2788: Java Runtime Environment Image Parsing Code buffer overflow vulnerability
     */
    javaicc_handler_init(bundle, inAPEToolsApplicationID);
    javaicc_handler_14_init(bundle, inAPEToolsApplicationID);
}

OSStatus APEBundleMessage(CFStringRef message, CFDataRef inData, CFDataRef *outData) {
    /*
     * APE messages are received on an independent thread, and any shared data access
     * must be synchronized.
     *
     * The only message currently supported is the kMoabPreferencesChanged notification.
     */

    if (CFStringCompare(message, kMoabPreferencesChanged, 0) == kCFCompareEqualTo) {
        /* Notify any patches that care about preference updates */
    }
    return noErr;
}
