/*
 * javaicc_handler.c
 * CVE-2007-2788 - Java Runtime Environment Image Parsing Code buffer overflow vulnerability
 *
 * Copyright (c) 2007 Landon Fuller <landonf@opendarwin.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Landon Fuller nor the names of any contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * SpProfileLoadFromBufferImp() is called with a ProfileData structure, and a pointer to the base address
 * of a buffer containing the raw profile data. The function parses out the tags, and this is where the
 * errant integer overflow occurs:
 *
 *      if (tagSize + tagOffset > profileSize) error;
 * 
 * This fails where tagSize + tagOffset > UINT_MAX
 *
 * As SpProfileLoadFromBufferImp() parses out the tags, it calls SpTagDirEntryAdd for each tag.
 * SpTagDirEntryAdd is passed a pointer (tagData), which is a pointer to the base address + the
 * tag offset, and tagSize, which is the actual tag data size.
 *
 * Note that we don't have access to the tag offset or base pointer here in SpTagDirEntry() -- thus,
 * we can't do an overflow-safe check of the tagSize + tagOffset. We can't easily replace SpProfileLoadFromBufferImp() and
 * fix it's check, either -- it's a fairly large function and there's a good chance of getting something wrong.
 *
 * Instead, we use thread-local data to set a context in a SpProfileLoadFromBufferImp(). The context contains
 * the base address, which we can use to derive the tagoffset, and then safely do the size check inside of
 * SpTagDirEntryAdd.
 * If this seems confusing, the code is probably more clear.
 *
 * Also, we implement a context stack, as SpProfileLoadFromBufferImp could be called recursively (I don't think so, though).
 */

#include <ApplicationEnhancer/ApplicationEnhancer.h>

#include <stdlib.h>
#include <mach-o/dyld.h>
#include <pthread.h>

#include "Moab.h"

#include <JavaVM/jni.h>

#include "javaicc_handler.h"

static const CFStringRef JREVersion = CFSTR("11.5.0");

/* This is returned if the tag is too long */
#define	ERR_INVALID_PROFILE 521

/* Thread local data */
typedef struct ProfileContext {
    uint32_t profileSize;
    char *baseAddr;
} ProfileContext;

static pthread_key_t profileContext;


/* Function pointer prototypes */
typedef int32_t (*func_SpTagDirEntryAdd)(void *profileData, uint32_t tagId, uint32_t dataSize, void *tagAddr);
typedef int32_t (*func_SpProfileLoadFromBufferImp)(void *profileData, char *baseAddr);
typedef uint32_t (*func_SpGetUInt32)(char **ptr);


/* Original Functions */
static func_SpTagDirEntryAdd            orig_SpTagDirEntryAdd           = NULL;
static func_SpProfileLoadFromBufferImp  orig_SpProfileLoadFromBufferImp = NULL;
static func_SpGetUInt32                 orig_SpGetUInt32                = NULL;


/* Replacement Functions */
static int32_t guard_SpTagDirEntryAdd (void *profileData, uint32_t tagId, uint32_t tagSize, char *tagAddr) {
    ProfileContext *context;

    /* Check if we're being called with a profile context */
    context = pthread_getspecific(profileContext);
    if (context != NULL) {
        /* Derive the tag offset from the baseAddr - tagAddr */
        uint32_t tagOffset = tagAddr - context->baseAddr;

        /* Do the comparison, uint64_t that won't overflow */
        if ((uint64_t)tagOffset + tagSize > context->profileSize) {
            apeprintf("[MOAB] Bad ICC tag data passed to Java; Likely attempt to exploit CVE-2007-2788.\n");
            apeprintf("[MOAB] Tag Offset: %u Tag Size: %u Profile Size: %u\n", tagOffset, tagSize, context->profileSize);
            return ERR_INVALID_PROFILE;
        }
    }

    return orig_SpTagDirEntryAdd(profileData, tagId, tagSize, tagAddr);
}


static int32_t guard_SpProfileLoadFromBufferImp (void *profileData, char *baseAddr) {
    ProfileContext context, *saved;
    char *ptr;
    int32_t retVal;

    if (baseAddr == NULL) {
        /* Supported, nothing to do */
        return orig_SpProfileLoadFromBufferImp(profileData, baseAddr);
    }

    /* Save the baseAddr, deserialize the profile size, and set our context */
    context.baseAddr = baseAddr;
    ptr = baseAddr;
    context.profileSize = orig_SpGetUInt32(&ptr);

    /* Save the old context */
    saved = pthread_getspecific(profileContext);
    if (pthread_setspecific(profileContext, &context) != 0) {
        // Shouldn't occur
        apeprintf("[MOAB] pthread_setspecific() failed in guard_SpProfileLoadFromBufferImp");
        return ERR_INVALID_PROFILE;
    }

    /* Call the original function */
    retVal = orig_SpProfileLoadFromBufferImp(profileData, baseAddr);

    /* Restore the old context, if any */
    if (pthread_setspecific(profileContext, saved) != 0) {
        // Shouldn't occur
        apeprintf("[MOAB] pthread_setspecific() failed in guard_SpProfileLoadFromBufferImp");
        return ERR_INVALID_PROFILE;
    }

    /* Done */
    return (retVal);
}


static void javaicc_dyld_image_add(const struct mach_header *header, intptr_t slide) {
    struct mach_header *jni_header;
    const char *imagePath;

    /*
     * Try to find the symbols from the libsuncmm jnilib.
     */
    if (orig_SpTagDirEntryAdd == NULL || orig_SpProfileLoadFromBufferImp == NULL || orig_SpGetUInt32 == NULL) {
        jni_header = MOABGetImageHeader(".*JavaVM.framework/Versions/1\\.5.*/Libraries/libsuncmm\\.jnilib$", &imagePath);
        if (jni_header) {
            if (!imagePath || !MOABMatchVersionForImageBundle(imagePath, 2, JREVersion)) {
                apeprintf("[MOAB] Skipping unsupported version of the Java runtime.\n");
                return;
            }

            /* Found the header, now find the symbols */
            if (orig_SpTagDirEntryAdd == NULL) {
                orig_SpTagDirEntryAdd = APEFindSymbol(jni_header, "_SpTagDirEntryAdd");

                if (!orig_SpTagDirEntryAdd) {
                    /* Symbol not found */
                    apeprintf("[MOAB] Failed to locate _SpTagDirEntryAdd symbol from Java libsuncmm \n");
                } else {
                    /* Symbol found, patch it */
                    orig_SpTagDirEntryAdd = APEPatchCreate(orig_SpTagDirEntryAdd, &guard_SpTagDirEntryAdd);
                }
            }

            if (orig_SpProfileLoadFromBufferImp == NULL) {
                orig_SpProfileLoadFromBufferImp = APEFindSymbol(jni_header, "_SpProfileLoadFromBufferImp");

                if (!orig_SpProfileLoadFromBufferImp) {
                    /* Symbol not found */
                    apeprintf("[MOAB] Failed to locate _SpProfileLoadFromBufferImp symbol from Java libsuncmm \n");
                } else {
                    /* Symbol found, patch it */
                    orig_SpProfileLoadFromBufferImp = APEPatchCreate(orig_SpProfileLoadFromBufferImp, &guard_SpProfileLoadFromBufferImp);
                }
            }
            
            if (orig_SpGetUInt32 == NULL) {
                orig_SpGetUInt32 = APEFindSymbol(jni_header, "_SpGetUInt32");

                if (!orig_SpGetUInt32)
                    apeprintf("[MOAB] Failed to locate _SpGetUInt32 symbol from Java libsuncmm \n");
            }            
        }
    }
}


void javaicc_handler_init(CFBundleRef bundle, CFStringRef inAPEToolsApplicationID) {
    /* Initialize our thread local storage */
    pthread_key_create(&profileContext, NULL);

    /* Register call back so we can watch for Java */
    _dyld_register_func_for_add_image(javaicc_dyld_image_add);
}